
# duration says how many SECONDS the animation lasts for
# id indicates the identity of multiple animated values
#   (i.e., allows a vector of animated values)
#   If "auto" then it depends on the number and size
#     of the elements being animated.  If there is
#     only one element, it is NULL.
# rep says how many times to repeat the animation
#   (TRUE means indefinitely;  FALSE means once)
# revert says whether to revert to the start value of the
#   animation upon completion

grid.animate <- function(x, ...,
                         duration=1, id="auto",
                         rep=FALSE, revert=FALSE) {
  UseMethod("grid.animate")
}

grid.animate.default <- function(x, ...,
                                 duration=1, id="auto",
                                 rep=FALSE, revert=FALSE) {
  stop("Animation not possible or not implemented for this type of object")
}

# May be possible to write a gpar method?

autoid <- function(id) {
  if (!is.numeric(id))
    if (id == "auto")
      TRUE
    else
      stop("Invalid id")
  else
    FALSE
}

grid.animate.unit <- function(x, ...,
                              duration=1, id="auto",
                              rep=FALSE, revert=FALSE) {
  if (missing(...))
    # Only one element to animate
    time.unit(x, duration,
              if (autoid(id)) NULL else id,
              rep, revert)
  else {
    if (autoid(id)) {
      temp <- list(...)
      n <- length(temp) + 1 # plus 1 because there's x too
      lengths <- c(unit.length(x),
                   unlist(lapply(temp, unit.length)))
      id <- rep(1:n, lengths)
    }
    time.unit(unit.c(x, unit.c(...)),
              duration, id, rep, revert)
  }
}

##############
# time.unit
##############

time.unit <- function(unit, duration=1, id=NULL,
                      rep=FALSE, revert=FALSE) {
  tu <- unit
  cl <- class(tu)
  attr(tu, "dur") <- duration
  attr(tu, "rep") <- rep
  attr(tu, "revert") <- revert
  attr(tu, "id") <- id
  class(tu) <- c("time.unit", cl)
  tu
}

# Convert a time unit to a normal unit
# Basically take the first value in each element of the time unit
# and produce a unit from them
timeUnitToUnit <- function(x) {
  id <- attr(x, "id")
  class(x) <- class(x)[-(match("time.unit", class(x)))]
  if (is.null(id))
    x[1]
  else
    x[match(unique(id), id)]
}

# Are two time units "identical"?
identical.time.unit <- function(x, y) {
  idx <- attr(x, "id")
  idy <- attr(y, "id")
  if (is.null(idx)) idx <- 1
  if (is.null(idy)) idy <- 1
  length(idx) == length(idy) &&
  attr(x, "dur") == attr(y, "dur") &&
  attr(x, "rep") == attr(y, "rep") &&
  attr(x, "revert") == attr(y, "revert") &&
  all(attr(x, "id") == attr(y, "id"))
}

#################
# Methods for time units (as.character, subsetting, arithmetic, ...)
#################

"[.time.unit" <- function(x, index, top=TRUE, ...) {
  id <- attr(x, "id")
  if (is.null(id)) {
    if (index > 1)
      stop("Invalid time unit index")
    result <- x
  } else {
    dur <- attr(x, "dur")
    rep <- attr(x, "rep")
    revert <- attr(x, "revert")
    # NOTE the *redefinition* of "index" here
    # This is deliberate to work with the behaviour of NextMethod
    index <- (1:length(id))[id == index]
    if (length(index) == 0)
      stop("Invalid time unit index")      
    result <- NextMethod()
    attr(result, "dur") <- dur
    attr(result, "rep") <- rep
    attr(result, "revert") <- revert
    attr(result, "id") <- id[index]
    # class(result) <- c("time.unit", class(result))
  }
  result
}  

# A time.unit has length 1
unit.length.time.unit <- function(unit) {
  id <- attr(unit, "id")
  if (is.null(id))
    1
  else
    length(unique(id))
}

as.character.time.unit <- function(x) {
  id <- attr(x, "id")
  if (is.null(id))
    paste("{", paste(NextMethod(), collapse=";"), "}",
          "(", attr(x, "dur"), "seconds)", sep="")
  else {
    temp <- NextMethod(x)
    paste("{", tapply(temp, id, paste, collapse=";"), "}",
          "(", attr(x, "dur"), "seconds)", sep="")
  }
}

WAS.Ops.time.unit <- function(x, y) {
  ok <- switch(.Generic, "+"=TRUE, "-"=TRUE, "*"=TRUE, FALSE)
  if (!ok)
    stop(paste("Operator", .Generic, "not meaningful for time units"))
  # Save time unit attributes
  id <- attr(x, "id")
  dur <- attr(x, "dur")
  rep <- attr(x, "rep")
  revert <- attr(x, "revert")
  if (.Generic=="*")
    # can only multiply a unit by a scalar
    if (nchar(.Method[1])) {
      if (nchar(.Method[2]))
        stop("Only one operand may be a time unit")
      else if (is.numeric(y)) {
        # Call NextMethod to get an appropriate unit object
        # then reattach the original time unit attributes
        result <- NextMethod()
        attr(result, "dur") <- dur
        attr(result, "rep") <- rep
        attr(result, "revert") <- revert
        attr(result, "id") <- id
      } else {
        stop("Non-time-unit operand must be numeric")
      } 
    } else {
      if (is.numeric(x)) {
        result <- NextMethod()
        attr(result, "dur") <- dur
        attr(result, "rep") <- rep
        attr(result, "revert") <- revert
        attr(result, "id") <- id
      } else {
        stop("Non-time-unit operand must be numeric")
      }
    }
  else
    # We are adding or subtracting units
    # Simple case is if both are time units
    if (nchar(.Method[1]) && nchar(.Method[2])) {
      # Only allowed for identical time units (for now, for simplicity)
      if (identical.time.unit(x, y)) {
        result <- NextMethod()
        attr(result, "dur") <- dur
        attr(result, "rep") <- rep
        attr(result, "revert") <- revert
        attr(result, "id") <- id
      } else {
        stop("Invalid to add or subtract non-identical time units")
      }
    } else if (nchar(.Method[1]) && is.unit(y)) {
      result <- unitPlusMinusTimeUnit(y, x, .Generic)
    } else if (nchar(.Method[2]) && is.unit(x)) {
      result <- unitPlusMinusTimeUnit(x, y, .Generic)
    } else {
      stop("Both operands must be units or derived from units")
    }
  result
}    

unitPlusMinusTimeUnit <- function(u, tu, op) {
  id <- attr(tu, "id")
  dur <- attr(tu, "dur")
  rep <- attr(tu, "rep")
  revert <- attr(tu, "revert")
  tunormal <- tu
  class(tunormal) <- class(tu)[-(match("time.unit", class(tu)))]
  nu <- unit.length(u)
  ntu <- unit.length(tu)
  if (is.null(id))
    tulengths <- unit.length(tunormal)
  else
    tulengths <- tapply(id, id, length)
  ldiff <- nu - ntu
  # If unit object is "longer" than time unit object
  # recycle time unit object
  if (ldiff > 0) {
    tulengths <- rep(tulengths, length=ntu + ldiff)
    tunormal <- unit.rep(tunormal, length=ntu + sum(ldiff*tulengths[1:ldiff]))
    id <- rep(id, length=ntu + sum(ldiff*tulengths[1:ldiff]))
  } else {
    u <- unit.rep(u, length=nu + ldiff)
  }
  z <- unit.rep(u, tulengths)
  result <- eval(call(op, z, tunormal))
  attr(result, "dur") <- dur
  attr(result, "rep") <- rep
  attr(result, "revert") <- revert
  attr(result, "id") <- id
  result
}

##############
# Hard to do generic device interface
# So must do something SVG-specific for now?
##############

# The approach here is based on reaching back up the call stack
# to determine the context in which the generic cx() was called.
# This is in contrast to the approach of extending the arguments
# passed to cx().
# The reasons for this choice are: (i) the arguments to cx() would
# have to be extended to a ridiculous extent and (ii) this approach
# allows animate.R to build on griddev.R with zero impact on the
# original design of griddev.R

methodName <- function(callstr) {
  # The calling functions are of the form:
  #   grobToDev.<something>(x, dev)
  dotindex <- match(".", callstr)
  parindex <- match("(", callstr)
  substr(callstr, dotindex, parindex - dotindex)
}

cx.time.unit <- function(x) {
  parentdev <- get("dev", env=sys.frame(sys.parent()))
  # Don't know the right general device interface for this sort
  # of thing yet, so special case for the SVG device
  if (is(parentdev, "svgDevice")) {
    xvalues <- grid.convertX(x, "native", valueOnly=TRUE)
    n <- unit.length(x)
    id <- attr(x, "id")
    # Determine the calling function in order to figure out
    # what to call the attribute being animated
    parentmethod <- sub("[(].+[)]$", "",
                        sub("^[^.]+[.]", "",
                            deparse(sys.call(sys.parent()))))
    print(parentmethod)
    # rect and circle are straightforward
    if (parentmethod %in% c("rect", "circle")) {
      attrib <- switch(parentmethod,
                       rect="x",
                       circle="cx")
      svgAnimateX(attrib, xvalues, id,
                  attr(x, "dur"), attr(x, "rep"),
                  attr(x, "revert"), parentdev@dev)
    # lines, segments, and polygon are trickier because we
    # need to combine x and y values into a "points" attribute
    } else if (parentmethod %in% c("lines", "segments", "polygon")) {
      attrib <- "points"
      parenty <- get("x", env=sys.frame(sys.parent()))$y
      yvalues <- grid.convertY(parenty, "native", valueOnly=TRUE)
      # Proceed IF 
      #   y is a time unit AND it has the same length AND
      #   duration AND repeat count as x
      #   OR
      #   y is a "normal" unit AND it has same unit.length
      if ((inherits(parenty, "time.unit") &&
           identical.time.unit(x, parenty))) 
        svgAnimatePoints(xvalues, yvalues, id,
                         attr(x, "dur"), attr(x, "rep"),
                         attr(x, "revert"), parentdev@dev)
      # If y is "normal", replicate yvalues to match xvalues
      else if (length(yvalues == n))
        svgAnimatePoints(xvalues, yvalues[id], id,
                         attr(x, "dur"), attr(x, "rep"),
                         attr(x, "revert"), parentdev@dev)
      else
        stop("Unit incompatible with time unit")
    # text is trickier too because we have to animate a transformation
    } else if (parentmethod == "text") {
      parenty <- get("x", env=sys.frame(sys.parent()))$y
      yvalues <- grid.convertY(parenty, "native", valueOnly=TRUE)
      # Proceed IF 
      #   y is a time unit AND it has the same length AND
      #   duration AND repeat count as x
      #   OR
      #   y is a "normal" unit AND it has length 1
      if ((inherits(parenty, "time.unit") &&
           identical.time.unit(x, parenty)) ||
          length(yvalues == 1))
        svgAnimateTranslation(xvalues, yvalues, id,
                              attr(x, "dur"), attr(x, "rep"),
                              attr(x, "revert"), parentdev@dev)
      else
        stop("Unit incompatible with time unit")
    # points are not implemented yet
    # viewports are not possible in the current setup(?)
    } else {
      warning("Time units not valid in this context")
    }
    # Return the first value of each time unit element in the animation
    grid.convertX(timeUnitToUnit(x), "native", valueOnly=TRUE)
  } else {
    # For other devices return the complete sequence of animated values
    NextMethod()
  }
}

cy.time.unit <- function(x) {
  parentdev <- get("dev", env=sys.frame(sys.parent()))
  # Don't know the right general device interface for this sort
  # of thing yet, so special case for the SVG device
  if (is(parentdev, "svgDevice")) {
    yvalues <- grid.convertY(x, "native", valueOnly=TRUE)
    n <- unit.length(x)
    id <- attr(x, "id")
    # Determine the calling function in order to figure out
    # what to call the attribute being animated
    parentmethod <- sub("[(].+[)]$", "",
                        sub("^[^.]+[.]", "",
                            deparse(sys.call(sys.parent()))))
    print(parentmethod)
    # rect and circle are straightforward
    if (parentmethod %in% c("rect", "circle")) {
      attrib <- switch(parentmethod,
                       rect="y",
                       circle="cy")
      svgAnimateY(attrib, yvalues, id,
                  attr(x, "dur"), attr(x, "rep"),
                  attr(x, "revert"), parentdev@dev)
    # lines, segments, and polygon are trickier because we
    # need to combine x and y values into a "points" attribute
    } else if (parentmethod %in% c("lines", "segments", "polygon")) {
      attrib <- "points"
      parentx <- get("x", env=sys.frame(sys.parent()))$x
      # IF x was a time unit then it will have already created
      # the animation so do nothing
      if (inherits(parentx, "time.unit")) {
        ;
      } else {
        xvalues <- grid.convertX(parentx, "native", valueOnly=TRUE)
        # Proceed IF 
        #   x is a "normal" unit AND it has same length as time unit
        if (length(xvalues == n))
          svgAnimatePoints(xvalues[id], yvalues, id,
                           attr(x, "dur"), attr(x, "rep"),
                           attr(x, "revert"), parentdev@dev)
        else
          stop("Unit incompatible with time unit")
      }
    # text is trickier too because we have to animate a transformation
    } else if (parentmethod == "text") {
      parentx <- get("x", env=sys.frame(sys.parent()))$x
      # IF x was a time unit then it will have already created
      # the animation so do nothing
      if (inherits(parentx, "time.unit")) {
        ;
      } else {
        xvalues <- grid.convertX(parentx, "native", valueOnly=TRUE)
        # Proceed IF 
        #   x is a "normal" unit AND it has length 1
        if (length(xvalues == 1))
        svgAnimateTranslation(xvalues, yvalues, id,
                              attr(x, "dur"), attr(x, "rep"),
                              attr(x, "revert"), parentdev@dev)
        else
          stop("Unit incompatible with time unit")
      }
    # points are not implemented yet
    # viewports are not possible in the current setup(?)
    } else {
      warning("Time units not valid in this context")
    }
    # Return the first value in the animation
    grid.convertY(timeUnitToUnit(x), "native", valueOnly=TRUE)
  } else {
    # For other devices return the complete sequence of animated values
    NextMethod()
  }
}

cw.time.unit <- function(x, dev) {
  parentdev <- get("dev", env=sys.frame(sys.parent()))
  # Don't know the right general device interface for this sort
  # of thing yet, so special case for the SVG device
  if (is(parentdev, "svgDevice")) {
    wvalues <- grid.convertWidth(x, "native", valueOnly=TRUE)
    n <- unit.length(x)
    id <- attr(x, "id")
    # Determine the calling function in order to figure out
    # what to call the attribute being animated
    parentmethod <- sub("[(].+[)]$", "",
                        sub("^[^.]+[.]", "",
                            deparse(sys.call(sys.parent()))))
    print(parentmethod)
    # rect is straightforward
    if (parentmethod %in% c("rect")) {
      attrib <- "width"
      svgAnimateWidth(attrib, wvalues, id,
                      attr(x, "dur"), attr(x, "rep"),
                      attr(x, "revert"), parentdev@dev)
    # points are not implemented yet
    # viewports are not possible in the current setup(?)
    } else {
      warning("Time units not valid in this context")
    }
    # Return the first value of each time unit element in the animation
    grid.convertWidth(timeUnitToUnit(x), "native", valueOnly=TRUE)
  } else {
    # For other devices return the complete sequence of animated values
    NextMethod()
  }
}

ch.time.unit <- function(x, dev) {
  parentdev <- get("dev", env=sys.frame(sys.parent()))
  # Don't know the right general device interface for this sort
  # of thing yet, so special case for the SVG device
  if (is(parentdev, "svgDevice")) {
    hvalues <- grid.convertHeight(x, "native", valueOnly=TRUE)
    n <- unit.length(x)
    id <- attr(x, "id")
    # Determine the calling function in order to figure out
    # what to call the attribute being animated
    parentmethod <- sub("[(].+[)]$", "",
                        sub("^[^.]+[.]", "",
                            deparse(sys.call(sys.parent()))))
    print(parentmethod)
    # rect is straightforward
    if (parentmethod %in% c("rect")) {
      attrib <- "height"
      svgAnimateHeight(attrib, hvalues, id,
                       attr(x, "dur"), attr(x, "rep"),
                       attr(x, "revert"), parentdev@dev)
    # points are not implemented yet
    # viewports are not possible in the current setup(?)
    } else {
      warning("Time units not valid in this context")
    }
    # Return the first value of each time unit element in the animation
    grid.convertHeight(timeUnitToUnit(x), "native", valueOnly=TRUE)
  } else {
    # For other devices return the complete sequence of animated values
    NextMethod()
  }
}

cd.time.unit <- function(x, dev) {
  parentdev <- get("dev", env=sys.frame(sys.parent()))
  # Don't know the right general device interface for this sort
  # of thing yet, so special case for the SVG device
  if (is(parentdev, "svgDevice")) {
    # Take the physically smaller value
    wvalues <- grid.convertWidth(x, "inches", valueOnly=TRUE)
    hvalues <- grid.convertHeight(x, "inches", valueOnly=TRUE)
    dvalues <- pmin(wvalues, hvalues)
    # For an SVG device, the "native" coordinates we are using
    # are "pixels" so are the same in each direction????
    # Hence doesn't matter whether I do a width or height conversion
    dvalues <- grid.convertWidth(unit(dvalues, "inches"), "native",
                                 valueOnly=TRUE)
    n <- unit.length(x)
    id <- attr(x, "id")
    # Determine the calling function in order to figure out
    # what to call the attribute being animated
    parentmethod <- sub("[(].+[)]$", "",
                        sub("^[^.]+[.]", "",
                            deparse(sys.call(sys.parent()))))
    print(parentmethod)
    # rect is straightforward
    if (parentmethod %in% c("circle")) {
      attrib <- "r"
      svgAnimateDim(attrib, dvalues, id,
                    attr(x, "dur"), attr(x, "rep"),
                    attr(x, "revert"), parentdev@dev)
    # points are not implemented yet
    # viewports are not possible in the current setup(?)
    } else {
      warning("Time units not valid in this context")
    }
    # Return the first value of each time unit element in the animation
    grid.convertWidth(timeUnitToUnit(x), "native", valueOnly=TRUE)
  } else {
    # For other devices return the complete sequence of animated values
    NextMethod()
  }
}

