
# Functions to take a grid grob and call appropriate
# functions from dev.R to produce output on a device

# Each function has to convert locations and dimenions
# into user coordinates THEN call the dev.R function

# Convert a unit object to a value in "native" units
cx <- function(x) {
  UseMethod("cx")
}

cy <- function(x) {
  UseMethod("cy")
}

cw <- function(x) {
  UseMethod("cw")
}

ch <- function(x) {
  UseMethod("ch")
}

cd <- function(x) {
  UseMethod("cd")
}

cx.unit <- function(x) {
  grid.convertX(x, "native", valueOnly=TRUE)
}

cy.unit <- function(x) {
  grid.convertY(x, "native", valueOnly=TRUE)
}

cw.unit <- function(x) {
  grid.convertWidth(x, "native", valueOnly=TRUE)
}

ch.unit <- function(x) {
  grid.convertHeight(x, "native", valueOnly=TRUE)
}

# Convert a "distance" (e.g., a circle radius)
cd.unit <- function(x) {
  pmin(grid.convertWidth(x, "native", valueOnly=TRUE),
       grid.convertHeight(x, "native", valueOnly=TRUE))
}
  
# Convert a gpar object to an svgStyle object
gparToDevPars <- function(gp) {
  if (length(gp) > 0)
    gp
  else
    list()
}

# Generate (left, bottom) from (x, y), (width, height), and justification
leftbottom <- function(x, y, width, height, just) {
  left <- switch(just[1],
                 left=x,
                 center=unit(grid.convertX(x, "native", valueOnly=TRUE) -
                   grid.convertWidth(0.5*width, "native", valueOnly=TRUE),
                   "native"),
                 centre=unit(grid.convertX(x, "native", valueOnly=TRUE) -
                   grid.convertWidth(0.5*width, "native", valueOnly=TRUE),
                   "native"),
                 right=unit(grid.convertX(x, "native", valueOnly=TRUE) -
                   grid.convertWidth(width, "native", valueOnly=TRUE),
                   "native"),
                 # If it's "bottom" or "top" just use "centre" for
                 # horizontal align
                 unit(grid.convertX(x, "native", valueOnly=TRUE) -
                   grid.convertWidth(0.5*width, "native", valueOnly=TRUE),
                   "native"))
  bottom <- switch(if (length(just) > 1) just[2] else just[1],
                   bottom=y,
                   center=unit(grid.convertY(y, "native", valueOnly=TRUE) -
                     grid.convertHeight(0.5*height, "native", valueOnly=TRUE),
                     "native"),
                   centre=unit(grid.convertY(y, "native", valueOnly=TRUE) -
                     grid.convertHeight(0.5*height, "native", valueOnly=TRUE),
                     "native"),
                   top=unit(grid.convertY(y, "native", valueOnly=TRUE) -
                     grid.convertHeight(height, "native", valueOnly=TRUE),
                     "native"),
                   # If it's "left" or "right" then use "centre" for
                   # vertical align
                   unit(grid.convertY(y, "native", valueOnly=TRUE) -
                     grid.convertHeight(0.5*height, "native", valueOnly=TRUE),
                     "native"))
  list(left=left, bottom=bottom)
}

# Generate hjust/vjust from just
justTohjust <- function(just) {
  if (is.na(match(just[1], c("left", "right"))))
    "centre"
  else
    just[1]
}

justTovjust <- function(just) {
  if (length(just) > 1) {
    just <- just[2]
  }
  if (is.na(match(just[1], c("top", "bottom"))))
    "centre"
  else
    just
}

# Grob to SVG
grobToDev <- function(x, dev) {
  UseMethod("grobToDev", x)
}

grobToDev.default <- function(x, dev) {
  stop("We shouldn't be here!")
}

grobToDev.lines <- function(x, dev) {
  if (!is.null(x$vp))
    gridToDev(x$vp, dev)
  devLines(cx(x$x), cy(x$y), gparToDevPars(x$gp), dev)
  if (!is.null(x$vp))
    gridToDev(1, dev)
}

# Any more efficient way of doing this?
grobToDev.segments <- function(x, dev) {
  if (!is.null(x$vp))
    gridToDev(x$vp, dev)
  nx0 <- unit.length(x$x0)
  nx1 <- unit.length(x$x1)
  ny0 <- unit.length(x$y0)
  ny1 <- unit.length(x$y1)
  n <- max(nx0, nx1, ny0, ny1)
  for (i in 1:n)
    devLines(cx(unit.c(x$x0[(i-1) %% nx0 + 1],
                       x$x1[(i-1) %% nx1 + 1])),
             cy(unit.c(x$y0[(i-1) %% ny0 + 1],
                       x$y1[(i-1) %% ny1 + 1])),
             gparToDevPars(x$gp), dev)
  if (!is.null(x$vp))
    gridToDev(1, dev)
}

grobToDev.polygon <- function(x, dev) {
  if (!is.null(x$vp))
    gridToDev(x$vp, dev)
  devPolygon(cx(x$x), cy(x$y), gparToDevPars(x$gp), dev)
  if (!is.null(x$vp))
    gridToDev(1, dev)
}

grobToDev.rect <- function(x, dev) {
  if (!is.null(x$vp))
    gridToDev(x$vp, dev)
  lb <- leftbottom(x$x, x$y, x$width, x$height, x$just)
  devRect(cx(lb$left), cy(lb$bottom),
          cw(x$width), ch(x$height),
          gparToDevPars(x$gp), dev)
  if (!is.null(x$vp))
    gridToDev(1, dev)
}

grobToDev.text <- function(x, dev) {
  if (!is.null(x$vp))
    gridToDev(x$vp, dev)
  devText(cx(x$x), cy(x$y), x$label,
          justTohjust(x$just), justTovjust(x$just),
          x$rot, gparToDevPars(x$gp), dev)
  if (!is.null(x$vp))
    gridToDev(1, dev)
}

grobToDev.circle <- function(x, dev) {
  if (!is.null(x$vp))
    gridToDev(x$vp, dev)
  devCircle(cx(x$x), cy(x$y), cd(x$r),
            gparToDevPars(x$gp), dev)
  if (!is.null(x$vp))
    gridToDev(1, dev)
}

# Quick fix for now
# Add device method eventually?
# Could get tricky to do all symbol types here ... (?)
grobToDev.points <- function(x, dev) {
  # ONLY pch = 1 or 3 handled
  if (!x$pch %in% c(1, 3))
    stop("Unsupported pch value")
  if (!is.null(x$vp))
    gridToDev(x$vp, dev)
  # length of x and y already checked in grid.points
  n <- unit.length(x$x)
  for (i in 1:n)
    if (x$pch == 1)
      devCircle(cx(x$x[i]),
                cy(x$y[i]),
                cd(0.5*x$size),
                gparToDevPars(x$gp), dev)
    else { # pch == 3
      deltax <- grid.convert(x$size, "native",
                             "x", "dimension",
                             "x", "location")
      devLines(cx(unit.c(x$x[i] - 0.5*deltax,
                         x$x[i] + 0.5*deltax)),
               cy(x$y[i]),
               gparToDevPars(x$gp), dev)
      deltay <- grid.convert(x$size, "native",
                             "y", "dimension",
                             "y", "location")
      devLines(cx(x$x[i]),
               cy(unit.c(x$y[i] - 0.5*deltay,
                         x$y[i] + 0.5*deltay)),
               gparToDevPars(x$gp), dev)      
    }
  if (!is.null(x$vp))
    gridToDev(1, dev)
}

grobToDev.collection <- function(x, dev) {
  if (!is.null(x$vp))
    gridToDev(x$vp, dev)
  devStartGroup(gparToDevPars(x$gp), dev)
  lapply(x$children, gridToDev, dev)
  devEndGroup(dev)
  if (!is.null(x$vp))
    gridToDev(1, dev)  
}

# Display list element to SVG
gridToDev <- function(x, dev) {
  UseMethod("gridToDev", x)
}

gridToDev.default <- function(x, dev) {
  if (!is.null(x)) {
    if (!is.numeric(x))
      stop("Something is very wrong here")
    devPopViewport(x, dev)
    # Actually pop the viewport(s) so that conversions, etc are correct
    pop.viewport(x, recording=FALSE)
  }
}

gridToDev.grob <- function(x, dev) {
  grobToDev(grid.get(x), dev)
}

gridToDev.viewport <- function(x, dev) {
  # Check whether viewport is positioned via layout
  if (!is.null(x$layout.pos.col) ||
      !is.null(x$layout.pos.row)) {
    pos <- layoutRegion(x$layout.pos.row, x$layout.pos.col)
    left <- pos$left
    bottom <- pos$bottom
    width <- pos$width
    height <- pos$height
  } else {
    # Positioning is via x, y, width, height
    lb <- leftbottom(x$x, x$y, x$width, x$height, x$just)
    left <- lb$left
    bottom <- lb$bottom
    width <- x$width
    height <- x$height
  }
  devPushViewport(cx(left), cy(bottom),
                  cw(width), ch(height),
                  x$xscale, x$yscale, gparToDevPars(x$gp), dev)
  # Actually push the viewport so that conversions, etc are correct
  push.viewport(x, recording=FALSE)
}

